% *************************************************************************
% ********************************** DOER *********************************
% A Dynamic and On-line Ensemble Regression for Changing Environments
% *************************************************************************
% Version 1.0 Nov/26/2015
%
% Developed in MATLAB R2010a
% 
% Author and programmer: Symone Gomes Soares
%
% e-mail: symonesoares@gmail.com 
%
% Homepage: http://sites.google.com/site/symonesoares/
%   
% Paper: SOARES, Symone Gomes; ARAÚJO, Rui. A Dynamic and On-line Ensemble 
%        Regression for Changing Environments. Expert Systems with
%        Applications, v. 42, p. 2935-2948, 2015. 
%        DOI: http://dx.doi.org/10.1016/j.eswa.2014.11.053
%
% *************************************************************************
% *************************************************************************
% Inputs of the Algorithm:
%
% - input(number of samples x number of input variables) : input matrix
% 
% - output(number of samples x output variable) : output matrix 
%   ps.: it accepts only one output variable
%
% - kfold   : number of folds for the k-fold cross-validation
%             Default: 10; 
% - sW      : window's size (sW >= kfold)
% - maxNHL  : maximum number of hidden nodes in OS-ELM (maxHL >= sW) 
%             Default: 20; 
% - Alpha   : factor to include a new model (0 < Alpha < 1);
%             Default: 0.04; 
% - ActF    : activation function for OS-ELM
%             Options: 'rbf'; 'sig'; 'sin'; 'hardlim';
%             Default: 'sig';
% - B       : maximum number of models in the ensemble (B > 1)
%             Default: 15;
% 
% *************************************************************************
% Outputs of the Algorithm:
%
% CC   : correlation coefficient between predicted outputs and real outputs
% MSE  : mean square errord between predicted outputs and real outputs
% RMSE : root mean square errord between predicted outputs and real outputs
% YR   : real outputs (testing data, i.e.\ Total of Samples minus sW)
% YP   : predicted outputs (testing data, i.e.\ Total of Samples minus sW)
% time : running time of the algorithm
%
% *************************************************************************
% *************************************************************************
% *************************************************************************

function [CC, MSE, RMSE, YR, YP, time] = fDOER(input, output, sW, maxNHL, Alpha, kfold, ActF, B)

    tStart = tic;
    
    %Setting some parameters 
    nS = size(input,1); nM = 1;  YP = [];  YR = [];
    if maxNHL > sW
         maxNHL = sW;
    end

    %Build the initial model
    XShifted    = input(1:sW,:);  %Assigning the inputs for the initial model
    YShifted    = output(1:sW,:); %Assigning the outputs for the initial model
    [Tmse, ~, ~] = KfoldCV_OSELM_ZeroMeanUnitVar(XShifted, YShifted, kfold, 1, [1 maxNHL], ActF);
    [~, Model(nM).nHL] = min(Tmse); 
    [Model(nM).mxcal,Model(nM).stdx, Model(nM).mycal, Model(nM).stdy, mcxcal, mcycal] = scaleZmean1Var(XShifted,YShifted);
    Model(nM).varx = (Model(nM).stdx).^2 ;
    Model(nM).vary = (Model(nM).stdy) ^2;
    [~, ~, Model(nM).ELM] = OSELM_Reg_Train(mcxcal, mcycal, Model(nM).nHL, ActF);
    Model(nM).w           = 1;
    Model(nM).S           = sW;
    Model(nM).life        = 0;
    Model(nM).e           = [];
    Model(nM).mse         = 0;
    nM                    = nM + 1;

    %Loop of samples
    for i = sW+1:nS

       xi = input(i,:);
       yi = output(i,:);
       yiModels  = zeros(1,size(Model,2)); yiEns = 0;
       for m = 1:size(Model,2)
          scxtest       = scale(xi, Model(m).mxcal, Model(m).stdx);  
          [~, yppred]   = OSELM_Reg_Test(scxtest, zeros(1,size(xi,1)), Model(m).ELM);
          yiModels(:,m) = rescale(yppred, Model(m).mycal, Model(m).stdy);
          yiEns         = yiEns + yiModels(:,m) * Model(m).w;
       end

       YP = [YP yiEns];
       YR = [YR yi];

       %Sliding the window
       XShifted(1,:) = [];
       YShifted(1,:) = [];
       XShifted = [XShifted; xi];
       YShifted = [YShifted; yi];

       for m = 1:size(Model,2)
            oldxmean       = Model(m).mxcal; 
            oldymean       = Model(m).mycal;
            Model(m).mxcal = Model(m).S / (Model(m).S + 1) * Model(m).mxcal + (1 / (Model(m).S + 1)) * xi;
            Model(m).mycal = Model(m).S / (Model(m).S + 1) * Model(m).mycal + (1 / (Model(m).S + 1)) * yi;     
            Model(m).varx  = ((Model(m).S - 1)/ (Model(m).S)) * Model(m).varx + (1 / Model(m).S) * ((xi - Model(m).mxcal).^2) + (Model(m).mxcal - oldxmean).^2;
            Model(m).vary  = ((Model(m).S - 1)/ (Model(m).S)) * Model(m).vary + (1 / Model(m).S) * ((yi - Model(m).mycal).^2) + (Model(m).mycal - oldymean).^2;
            Model(m).stdx  = sqrt(Model(m).varx);
            Model(m).stdy  = sqrt(Model(m).vary);
            Model(m).S     = Model(m).S + 1;
            xiS            = scale(xi, Model(m).mxcal, Model(m).stdx);
            yiS            = scale(yi, Model(m).mycal, Model(m).stdy);
            [Model(m).ELM, ~] = OSELM_Reg_Retrain(Model(m).ELM, xiS, yiS);
       end

       %Updating some parameters
       worstModel = -inf; M = [];
       for k = 1:size(Model,2)
           Model(k).life  = Model(k).life + 1;
           eNow          = (yi - yiModels(1,k))^2;
           Model(k).e    = [Model(k).e eNow];
           if  Model(k).life <= sW
               Model(k).mse = ((Model(k).life - 1) / Model(k).life) * Model(k).mse + (1 / Model(k).life) * eNow;
           else 
              Model(k).mse   = Model(k).mse + eNow/sW - Model(k).e(:,1)/sW;
               Model(k).e(:,1) = [];
           end
           if Model(k).mse > worstModel
               worstModel = Model(k).mse; ix = k;
           end
           M = [M Model(k).mse];
       end

       %Updating the weights
       sumW = 0;
       for m = 1:size(Model,2)
           Model(m).w = exp(-(Model(m).mse - median(M))/(median(M)));
           sumW = sumW + Model(m).w;
       end

       %Generating a new model
       if abs((yiEns - yi)/yi) > Alpha 
            [Tmse, ~, ~]              = KfoldCV_OSELM_ZeroMeanUnitVar(XShifted, YShifted, kfold, 1, [1 maxNHL], ActF);
            [~, Model(nM).nHL]        = min(Tmse);
            [Model(nM).mxcal,Model(nM).stdx, Model(nM).mycal, Model(nM).stdy, mcxcal, mcycal] = scaleZmean1Var(XShifted,YShifted);
            Model(nM).varx = (Model(nM).stdx).^2 ;
            Model(nM).vary = (Model(nM).stdy) ^2;
            [~, ~, Model(nM).ELM]     = OSELM_Reg_Train(mcxcal, mcycal, Model(nM).nHL, ActF);
            Model(nM).w               = 1; 
            Model(nM).S               = sW; 
            Model(nM).e               = [];
            Model(nM).mse             = 0; 
            Model(nM).life            = 0; 
            sumW                      = sumW + 1;
            nM = nM + 1;
       end

       if size(Model,2) > B
          sumW = sumW - Model(ix).w;
          Model(ix) = []; nM = nM - 1;
       end

       %Normalizing the weights
       for m = 1:size(Model,2)
           Model(m).w = Model(m).w/sumW;
       end

    end
    
    time = toc(tStart);
    CC   = corr2(YR,YP);
    MSE  = mse(YP - YR);
    RMSE = mse(YP - YR)^1/2;
    
    fprintf('\n\n\nCorrelation coefficient: %f \n', CC);
    fprintf('Mean Squared Error:      %f \n', MSE);
    fprintf('Root Mean Squared Error: %f \n', RMSE);
    
end
